// TicTacToeServerExt.java
// Select your personal session ID! (same for player)

import ch.aplu.tcp.*;
import java.util.*;
import ch.aplu.util.Console;

public class TicTacToeServerExt extends TcpBridge
{
  private static String VERSION = "1.2";
  private static String sessionID = "TicTacToeExt_8arewf7%&%2134";
  private static final String bridgeName = "TicTacToeServer";
  private static Console c;
  private Hashtable<String, RoomState> rooms =
    new Hashtable<String, RoomState>();

  public TicTacToeServerExt()
  {
    super(sessionID, bridgeName);
    c = Console.init();
    c.setTitle("TicTacToeServerExt Version " + VERSION);
    addTcpBridgeListener(new TcpBridgeAdapter()
    {
      public void pipeRequest(String source, String destination, int[] data)
      {
        showPipeRequest(source, data);
        RoomState roomState;
        ArrayList<String> players;

        switch (data[0])
        {
          case Command.REQUEST_GAME_ENTRY:
            String roomName = TcpTools.intAryToString(data, 1);
            c.println("Player " + source + " requests entry in room " + roomName);
            if (rooms.get(roomName) == null)  // No player in room
            {
              c.println("First player " + source + " entered in room " + roomName);
              roomState = new RoomState();
              rooms.put(roomName, roomState);
              roomState.players.add(source);
              send(bridgeName, source, Command.TEAMMATE_WAIT);
              return;
            }
            roomState = rooms.get(roomName);
            players = roomState.players;
            if (players.size() == 2)
            {
              c.println("Room: " + roomName + " -- Too many players. Game in process.");
              send(bridgeName, source, Command.IN_PROGRESS);
              return;
            }
            if (players.size() == 1)
            {
              c.println("Room: " + roomName + " -- Second player " + source + " entered.");
              players.add(source);
              // We transmit names of players
              String player0 = players.get(0);
              String player1 = players.get(1);
              c.println("Room: " + roomName + " -- Game starting");
              c.print("Players in game: ");
              c.print(player0 + " (entered first); ");
              c.println(player1 + " (entered second)");

              String text =
                TcpTools.stripNickname(player0, false)
                + ">"
                + TcpTools.stripNickname(player1, false);
              sendCommand(bridgeName, player0, Command.PLAYER_NAMES,
                TcpTools.stringToIntAry(text));

              text = TcpTools.stripNickname(player1, false)
                + ">"
                + TcpTools.stripNickname(player0, false);
              sendCommand(bridgeName, player1, Command.PLAYER_NAMES,
                TcpTools.stringToIntAry(text));

              roomState.startingGrid[0] = player0;
              roomState.startingGrid[1] = player1;
              initGame(roomName, player0, player1);
            }
            break;

          case Command.MOVE:
            roomName = getRoomName(source);
            roomState = rooms.get(roomName);
            String teammate = getTeammate(source);
            char[][] gameState = roomState.gameState;
            char value = data[3] == 0 ? 'O' : 'X';
            gameState[data[1]][data[2]] = value;
            send("", teammate, data);

            char chr = checkBoard(roomName);
            if (chr == ' ')  // still in progress
            {
              if (isBoardFull(roomName))
              {
                c.println("Room: " + roomName + " -- Game over. It's a draw");
                send("", source, Command.TIE);
                send("", teammate, Command.TIE);
                // Exchange starting grid
                String[] startingGrid = roomState.startingGrid;
                String temp = startingGrid[1];
                startingGrid[1] = startingGrid[0];
                startingGrid[0] = temp;
              }
            }
            else
            {
              String winner = (chr == value) ? source : teammate;
              String looser = (chr != value) ? source : teammate;
              c.println("Room: " + roomName + " -- Game over. Winner: " + winner + ". Looser: " + looser);
              send("", winner, Command.WON);
              send("", looser, Command.LOST);
              String[] startingGrid = roomState.startingGrid;
              startingGrid[0] = looser;  // Looser makes first move
              startingGrid[1] = winner;
            }
            break;

          case Command.RESTART:
            roomName = getRoomName(source);
            roomState = rooms.get(roomName);
            teammate = getTeammate(source);
            roomState.restartRequest++;
            if (roomState.restartRequest == 2)
            {
              players = roomState.players;
              initGame(roomName, players.get(0), players.get(1));
            }
            break;
        }
      }

      public void notifyAgentConnection(String agentName, boolean connected)
      {
        if (!connected)
          removePlayer(agentName);
      }
    });

    c.print("Connecting to relay...");
    if (connectToRelay().size() > 1)
    {
      c.println("Connected.\nBut game room ID in use.\nTerminating now...");
      TcpTools.delay(4000);
      System.exit(0);
    }

    c.println("done.\nWaiting for players...");
  }

  private void removePlayer(String agentName)
  {
    Enumeration e = rooms.keys();
    while (e.hasMoreElements())
    {
      String roomName = (String)e.nextElement();
      ArrayList<String> players = rooms.get(roomName).players;
      int nb = players.size();

      if (nb == 1)  // One waiting player
      {
        rooms.remove(roomName);
        c.println("Room: " + roomName + " -- Player " + agentName
          + " and room removed.");
        return;
      }

      if (nb == 2)  // Game in progress
      {
        for (int i = 0; i < nb; i++)
        {
          if (players.get(i).equals(agentName))
          {
            send(bridgeName, players.get((i + 1) % 2), Command.TEAMMATE_QUIT);
            players.remove(agentName);
            c.println("Room: " + roomName + " -- Player " + agentName
              + " removed.");
            return;
          }
        }
      }
    }
  }

  private boolean isBoardFull(String roomName)
  {
    char[][] gameState = rooms.get(roomName).gameState;
    int nb = 0;
    for (int i = 0; i < 3; i++)
      for (int k = 0; k < 3; k++)
        if (gameState[i][k] != ' ')
          nb++;
    return nb == 9;
  }

  private char checkBoard(String roomName)
  // Returns 'X' if XXX found, 'O' if OOO found, else ' '
  {
    char[][] gameState = rooms.get(roomName).gameState;

    // Convert gameState into string pattern
    StringBuffer pattern = new StringBuffer();
    // Horizontal
    for (int i = 0; i < 3; i++)
    {
      for (int k = 0; k < 3; k++)
      {
        pattern.append(gameState[i][k]);
      }
      pattern.append(',');  // Separator
    }
    // Vertical
    for (int k = 0; k < 3; k++)
    {
      for (int i = 0; i < 3; i++)
      {
        pattern.append(gameState[i][k]);
      }
      pattern.append(',');
    }
    // Diagonal
    for (int i = 0; i < 3; i++)
      pattern.append(gameState[i][i]);
    pattern.append(',');
    for (int i = 0; i < 3; i++)
      pattern.append(gameState[i][2 - i]);

    if (pattern.toString().contains("XXX"))
      return 'X';
    if (pattern.toString().contains("OOO"))
      return 'O';
    return ' ';
  }

  private void initGame(String roomName, String player0, String player1)
  {
    RoomState roomState = rooms.get(roomName);
    String[] startingGrid = roomState.startingGrid;
    char[][] gameState = roomState.gameState;
    for (int i = 0; i < 3; i++)
      for (int k = 0; k < 3; k++)
        gameState[i][k] = ' ';
    roomState.restartRequest = 0;
    send("", player0, Command.MARK_X);
    send("", player1, Command.MARK_O);
    // Send first to 'second' to make him ready to receive move requests
    send("", startingGrid[1], Command.START_SECOND);
    TcpTools.delay(2000);
    send("", startingGrid[0], Command.START_FIRST);
  }

  private String getTeammate(String player)
  // Returns empty string if player is not found or if he has no teammate
  {
    Enumeration e = rooms.keys();
    while (e.hasMoreElements())
    {
      String roomName = (String)e.nextElement();
      ArrayList<String> players = rooms.get(roomName).players;
      if (players.size() < 2)  // Only one player in room
        continue;
      for (int i = 0; i < 2; i++)
      {
        if (players.get(i).equals(player))
          return players.get((i + 1) % 2);
      }
    }
    return "";
  }

  private String getRoomName(String player)
  // Returns empty string if player is not found in any room
  {
    Enumeration e = rooms.keys();
    while (e.hasMoreElements())
    {
      String roomName = (String)e.nextElement();
      ArrayList<String> players = rooms.get(roomName).players;
      for (int i = 0; i < players.size(); i++)
      {
        if (players.get(i).equals(player))
          return roomName;
      }
    }
    return "";
  }

  private void showPipeRequest(String source, int[] data)
  {
    c.print("pipeRequest from: " + source);
    c.print(". Data: ");
    for (Integer k : data)
      c.print(k + " ");
    c.println();
  }

  public static void main(String[] args)
  {
    new TicTacToeServerExt();
  }
}
